/*----------------------------------------------------------------------------
  Simple class support routines such as list manipulation.
----------------------------------------------------------------------------*/
#include "atddr.h"

struct _utFunctionTable atMportFunctionTable = {utSymNull, atMportDestroy};
utSym atMportClassName;
atMport _atFirstFreeMport, _atTempMport;
utHeapRef _atMportHeap;
void (*atContextDestructorHook)(atContext);
atContext _atFirstFreeContext, _atTempContext;
utHeapRef _atContextHeap;
void (*atAttrDestructorHook)(atAttr);
atAttr _atFirstFreeAttr, _atTempAttr;
utHeapRef _atAttrHeap;
struct _utFunctionTable atPortFunctionTable = {utSymNull, atPortDestroy};
utSym atPortClassName;
atPort _atFirstFreePort, _atTempPort;
utHeapRef _atPortHeap;
void (*atRootDestructorHook)(atRoot);
atRoot _atFirstFreeRoot, _atTempRoot;
utHeapRef _atRootHeap;
void (*atDeviceDestructorHook)(atDevice);
atDevice _atFirstFreeDevice, _atTempDevice;
utHeapRef _atDeviceHeap;
void (*atPinDestructorHook)(atPin);
atPin _atFirstFreePin, _atTempPin;
utHeapRef _atPinHeap;

/*------------------------------------------------------------------------
  Allocate a new memory block for the Mport.
------------------------------------------------------------------------*/
static utBlockRef buildMportBlock(void)
{
    utBlockRef block = utcBlock();
    atMport Mport;
    U16 x, sMport;

    utaHeapBlock(_atMportHeap, block);
    sMport = UTBLOCKSIZE/sizeof(struct _atMport);
    _atFirstFreeMport = (atMport)utgBlockMem(block);
    Mport = _atFirstFreeMport;
    for(x = 0; x < sMport; x++) {
        _atMportNextFree(Mport) = Mport + 1;
        Mport++;
    }
    _atMportNextFree(--Mport) = atMportNull;
    return block;
}

/*----------------------------------------------------------------------------
  Allocate and initialize a new Mport.
----------------------------------------------------------------------------*/
atMport atMportAlloc(dbMport super)
{
    atMport Mport;

    if(_atFirstFreeMport == atMportNull) {
        buildMportBlock();
    }
    Mport = _atFirstFreeMport;
    _atFirstFreeMport = _atMportNextFree(Mport);
    memset((void *)Mport, 0, sizeof(struct _atMport));
    Mport->_extension.next = dbMportGetfirstExtension(super);
    dbMportSetfirstExtension(super, &Mport->_extension);
    Mport->_extension.functionTable = &atMportFunctionTable;
    Mport->_super = super;
    return Mport;
}

/*----------------------------------------------------------------------------
  Free a Mport.
----------------------------------------------------------------------------*/
void atMportFree(
    atMport Mport)
{
    dbMportSetfirstExtension(atMportSuper(Mport), Mport->_extension.next);
    _atMportNextFree(Mport) = _atFirstFreeMport;
    _atFirstFreeMport = Mport;
}

/*----------------------------------------------------------------------------
  Destroy a Mport.
----------------------------------------------------------------------------*/
void atMportDestroy(
    atMport Mport)
{
    atMportFree(Mport);
}

/*----------------------------------------------------------------------------
  Add the Attr to the tail of the list on the Context.
----------------------------------------------------------------------------*/
void atContextInsertAttr(
    atContext Context,
    atAttr _Attr)
{
    utInsertHtblEntry(atContextGetAttrs(Context), &(_Attr)->h);
    atAttrSetContext(_Attr, Context);
}

/*----------------------------------------------------------------------------
  Remove the Attr from the list on the Context.
----------------------------------------------------------------------------*/
void atContextRemoveAttr(
    atContext Context,
    atAttr _Attr)
{
    utDeleteHtblEntry(&(Context)->Attrs, &(_Attr)->h);
    atAttrSetContext(_Attr, atContextNull);
}

/*----------------------------------------------------------------------------
  Append the Attr to the tail of the list on the Context.
----------------------------------------------------------------------------*/
void atContextAppendAttr(
    atContext Context,
    atAttr _Attr)
{
    utAppendHtblEntry(atContextGetAttrs(Context), &(_Attr)->h);
    atAttrSetContext(_Attr, Context);
}

/*------------------------------------------------------------------------
  Allocate a new memory block for the Context.
------------------------------------------------------------------------*/
static utBlockRef buildContextBlock(void)
{
    utBlockRef block = utcBlock();
    atContext Context;
    U16 x, sContext;

    utaHeapBlock(_atContextHeap, block);
    sContext = UTBLOCKSIZE/sizeof(struct _atContext);
    _atFirstFreeContext = (atContext)utgBlockMem(block);
    Context = _atFirstFreeContext;
    for(x = 0; x < sContext; x++) {
        _atContextNextFree(Context) = Context + 1;
        Context++;
    }
    _atContextNextFree(--Context) = atContextNull;
    return block;
}

/*----------------------------------------------------------------------------
  Allocate and initialize a new Context.
----------------------------------------------------------------------------*/
atContext atContextAlloc(void)
{
    atContext Context;

    if(_atFirstFreeContext == atContextNull) {
        buildContextBlock();
    }
    Context = _atFirstFreeContext;
    _atFirstFreeContext = _atContextNextFree(Context);
    memset((void *)Context, 0, sizeof(struct _atContext));
    utInitHtbl(&Context->Attrs);
    return Context;
}

/*----------------------------------------------------------------------------
  Free a Context.
----------------------------------------------------------------------------*/
void atContextFree(
    atContext Context)
{
    utdArray(utfHtblArray(atContextGetAttrs(Context)));
    _atContextNextFree(Context) = _atFirstFreeContext;
    _atFirstFreeContext = Context;
}

/*----------------------------------------------------------------------------
  Destroy a Context.
----------------------------------------------------------------------------*/
void atContextDestroy(
    atContext Context)
{
    if(atContextDestructorHook != NULL) {
        atContextDestructorHook(Context);
    }
    {
        atAttr _Attr;
        atSafeForeachContextAttr(Context, _Attr) {
            atAttrDestroy(_Attr);
        } atEndSafeForeachContextAttr;
    }
    atContextFree(Context);
}

/*------------------------------------------------------------------------
  Allocate a new memory block for the Attr.
------------------------------------------------------------------------*/
static utBlockRef buildAttrBlock(void)
{
    utBlockRef block = utcBlock();
    atAttr Attr;
    U16 x, sAttr;

    utaHeapBlock(_atAttrHeap, block);
    sAttr = UTBLOCKSIZE/sizeof(struct _atAttr);
    _atFirstFreeAttr = (atAttr)utgBlockMem(block);
    Attr = _atFirstFreeAttr;
    for(x = 0; x < sAttr; x++) {
        _atAttrNextFree(Attr) = Attr + 1;
        Attr++;
    }
    _atAttrNextFree(--Attr) = atAttrNull;
    return block;
}

/*----------------------------------------------------------------------------
  Allocate and initialize a new Attr.
----------------------------------------------------------------------------*/
atAttr _atAttrAlloc(void)
{
    atAttr Attr;

    if(_atFirstFreeAttr == atAttrNull) {
        buildAttrBlock();
    }
    Attr = _atFirstFreeAttr;
    _atFirstFreeAttr = _atAttrNextFree(Attr);
    memset((void *)Attr, 0, sizeof(struct _atAttr));
    return Attr;
}

/*----------------------------------------------------------------------------
  Free a Attr.
----------------------------------------------------------------------------*/
void _atAttrFree(
    atAttr Attr)
{
    _atAttrNextFree(Attr) = _atFirstFreeAttr;
    _atFirstFreeAttr = Attr;
}

/*----------------------------------------------------------------------------
  Destroy a Attr.
----------------------------------------------------------------------------*/
void atAttrDestroy(
    atAttr Attr)
{
    if(atAttrDestructorHook != NULL) {
        atAttrDestructorHook(Attr);
    }
    {
        atContext parentContext = atAttrGetContext(Attr);
        if(parentContext != atContextNull) {
            atContextRemoveAttr(parentContext, Attr);
        }
    }
    atAttrFree(Attr);
}

/*------------------------------------------------------------------------
  Allocate a new memory block for the Port.
------------------------------------------------------------------------*/
static utBlockRef buildPortBlock(void)
{
    utBlockRef block = utcBlock();
    atPort Port;
    U16 x, sPort;

    utaHeapBlock(_atPortHeap, block);
    sPort = UTBLOCKSIZE/sizeof(struct _atPort);
    _atFirstFreePort = (atPort)utgBlockMem(block);
    Port = _atFirstFreePort;
    for(x = 0; x < sPort; x++) {
        _atPortNextFree(Port) = Port + 1;
        Port++;
    }
    _atPortNextFree(--Port) = atPortNull;
    return block;
}

/*----------------------------------------------------------------------------
  Allocate and initialize a new Port.
----------------------------------------------------------------------------*/
atPort atPortAlloc(dbPort super)
{
    atPort Port;

    if(_atFirstFreePort == atPortNull) {
        buildPortBlock();
    }
    Port = _atFirstFreePort;
    _atFirstFreePort = _atPortNextFree(Port);
    memset((void *)Port, 0, sizeof(struct _atPort));
    Port->_extension.next = dbPortGetfirstExtension(super);
    dbPortSetfirstExtension(super, &Port->_extension);
    Port->_extension.functionTable = &atPortFunctionTable;
    Port->_super = super;
    return Port;
}

/*----------------------------------------------------------------------------
  Free a Port.
----------------------------------------------------------------------------*/
void atPortFree(
    atPort Port)
{
    dbPortSetfirstExtension(atPortSuper(Port), Port->_extension.next);
    _atPortNextFree(Port) = _atFirstFreePort;
    _atFirstFreePort = Port;
}

/*----------------------------------------------------------------------------
  Destroy a Port.
----------------------------------------------------------------------------*/
void atPortDestroy(
    atPort Port)
{
    atPortFree(Port);
}

/*----------------------------------------------------------------------------
  Add the Device to the tail of the list on the Root.
----------------------------------------------------------------------------*/
void atRootInsertDevice(
    atRoot Root,
    atDevice _Device)
{
    utInsertHtblEntry(atRootGetDevices(Root), &(_Device)->h);
    atDeviceSetRoot(_Device, Root);
}

/*----------------------------------------------------------------------------
  Remove the Device from the list on the Root.
----------------------------------------------------------------------------*/
void atRootRemoveDevice(
    atRoot Root,
    atDevice _Device)
{
    utDeleteHtblEntry(&(Root)->Devices, &(_Device)->h);
    atDeviceSetRoot(_Device, atRootNull);
}

/*----------------------------------------------------------------------------
  Append the Device to the tail of the list on the Root.
----------------------------------------------------------------------------*/
void atRootAppendDevice(
    atRoot Root,
    atDevice _Device)
{
    utAppendHtblEntry(atRootGetDevices(Root), &(_Device)->h);
    atDeviceSetRoot(_Device, Root);
}

/*------------------------------------------------------------------------
  Allocate a new memory block for the Root.
------------------------------------------------------------------------*/
static utBlockRef buildRootBlock(void)
{
    utBlockRef block = utcBlock();
    atRoot Root;
    U16 x, sRoot;

    utaHeapBlock(_atRootHeap, block);
    sRoot = UTBLOCKSIZE/sizeof(struct _atRoot);
    _atFirstFreeRoot = (atRoot)utgBlockMem(block);
    Root = _atFirstFreeRoot;
    for(x = 0; x < sRoot; x++) {
        _atRootNextFree(Root) = Root + 1;
        Root++;
    }
    _atRootNextFree(--Root) = atRootNull;
    return block;
}

/*----------------------------------------------------------------------------
  Allocate and initialize a new Root.
----------------------------------------------------------------------------*/
atRoot atRootAlloc(void)
{
    atRoot Root;

    if(_atFirstFreeRoot == atRootNull) {
        buildRootBlock();
    }
    Root = _atFirstFreeRoot;
    _atFirstFreeRoot = _atRootNextFree(Root);
    memset((void *)Root, 0, sizeof(struct _atRoot));
    utInitHtbl(&Root->Devices);
    return Root;
}

/*----------------------------------------------------------------------------
  Free a Root.
----------------------------------------------------------------------------*/
void atRootFree(
    atRoot Root)
{
    utdArray(utfHtblArray(atRootGetDevices(Root)));
    _atRootNextFree(Root) = _atFirstFreeRoot;
    _atFirstFreeRoot = Root;
}

/*----------------------------------------------------------------------------
  Destroy a Root.
----------------------------------------------------------------------------*/
void atRootDestroy(
    atRoot Root)
{
    if(atRootDestructorHook != NULL) {
        atRootDestructorHook(Root);
    }
    {
        atDevice _Device;
        atSafeForeachRootDevice(Root, _Device) {
            atDeviceDestroy(_Device);
        } atEndSafeForeachRootDevice;
    }
    atRootFree(Root);
}

/*----------------------------------------------------------------------------
  Add the Pin to the tail of the list on the Device.
----------------------------------------------------------------------------*/
void atDeviceInsertPin(
    atDevice Device,
    atPin _Pin)
{
    utInsertHtblEntry(atDeviceGetPins(Device), &(_Pin)->h);
    atPinSetDevice(_Pin, Device);
}

/*----------------------------------------------------------------------------
  Remove the Pin from the list on the Device.
----------------------------------------------------------------------------*/
void atDeviceRemovePin(
    atDevice Device,
    atPin _Pin)
{
    utDeleteHtblEntry(&(Device)->Pins, &(_Pin)->h);
    atPinSetDevice(_Pin, atDeviceNull);
}

/*----------------------------------------------------------------------------
  Append the Pin to the tail of the list on the Device.
----------------------------------------------------------------------------*/
void atDeviceAppendPin(
    atDevice Device,
    atPin _Pin)
{
    utAppendHtblEntry(atDeviceGetPins(Device), &(_Pin)->h);
    atPinSetDevice(_Pin, Device);
}

/*------------------------------------------------------------------------
  Allocate a new memory block for the Device.
------------------------------------------------------------------------*/
static utBlockRef buildDeviceBlock(void)
{
    utBlockRef block = utcBlock();
    atDevice Device;
    U16 x, sDevice;

    utaHeapBlock(_atDeviceHeap, block);
    sDevice = UTBLOCKSIZE/sizeof(struct _atDevice);
    _atFirstFreeDevice = (atDevice)utgBlockMem(block);
    Device = _atFirstFreeDevice;
    for(x = 0; x < sDevice; x++) {
        _atDeviceNextFree(Device) = Device + 1;
        Device++;
    }
    _atDeviceNextFree(--Device) = atDeviceNull;
    return block;
}

/*----------------------------------------------------------------------------
  Allocate and initialize a new Device.
----------------------------------------------------------------------------*/
atDevice atDeviceAlloc(void)
{
    atDevice Device;

    if(_atFirstFreeDevice == atDeviceNull) {
        buildDeviceBlock();
    }
    Device = _atFirstFreeDevice;
    _atFirstFreeDevice = _atDeviceNextFree(Device);
    memset((void *)Device, 0, sizeof(struct _atDevice));
    utInitHtbl(&Device->Pins);
    return Device;
}

/*----------------------------------------------------------------------------
  Free a Device.
----------------------------------------------------------------------------*/
void atDeviceFree(
    atDevice Device)
{
    utdArray(utfHtblArray(atDeviceGetPins(Device)));
    _atDeviceNextFree(Device) = _atFirstFreeDevice;
    _atFirstFreeDevice = Device;
}

/*----------------------------------------------------------------------------
  Destroy a Device.
----------------------------------------------------------------------------*/
void atDeviceDestroy(
    atDevice Device)
{
    if(atDeviceDestructorHook != NULL) {
        atDeviceDestructorHook(Device);
    }
    {
        atPin _Pin;
        atSafeForeachDevicePin(Device, _Pin) {
            atPinDestroy(_Pin);
        } atEndSafeForeachDevicePin;
    }
    {
        atRoot parentRoot = atDeviceGetRoot(Device);
        if(parentRoot != atRootNull) {
            atRootRemoveDevice(parentRoot, Device);
        }
    }
    atDeviceFree(Device);
}

/*------------------------------------------------------------------------
  Allocate a new memory block for the Pin.
------------------------------------------------------------------------*/
static utBlockRef buildPinBlock(void)
{
    utBlockRef block = utcBlock();
    atPin Pin;
    U16 x, sPin;

    utaHeapBlock(_atPinHeap, block);
    sPin = UTBLOCKSIZE/sizeof(struct _atPin);
    _atFirstFreePin = (atPin)utgBlockMem(block);
    Pin = _atFirstFreePin;
    for(x = 0; x < sPin; x++) {
        _atPinNextFree(Pin) = Pin + 1;
        Pin++;
    }
    _atPinNextFree(--Pin) = atPinNull;
    return block;
}

/*----------------------------------------------------------------------------
  Allocate and initialize a new Pin.
----------------------------------------------------------------------------*/
atPin _atPinAlloc(void)
{
    atPin Pin;

    if(_atFirstFreePin == atPinNull) {
        buildPinBlock();
    }
    Pin = _atFirstFreePin;
    _atFirstFreePin = _atPinNextFree(Pin);
    memset((void *)Pin, 0, sizeof(struct _atPin));
    return Pin;
}

/*----------------------------------------------------------------------------
  Free a Pin.
----------------------------------------------------------------------------*/
void _atPinFree(
    atPin Pin)
{
    _atPinNextFree(Pin) = _atFirstFreePin;
    _atFirstFreePin = Pin;
}

/*----------------------------------------------------------------------------
  Destroy a Pin.
----------------------------------------------------------------------------*/
void atPinDestroy(
    atPin Pin)
{
    if(atPinDestructorHook != NULL) {
        atPinDestructorHook(Pin);
    }
    {
        atDevice parentDevice = atPinGetDevice(Pin);
        if(parentDevice != atDeviceNull) {
            atDeviceRemovePin(parentDevice, Pin);
        }
    }
    atPinFree(Pin);
}

/*----------------------------------------------------------------------------
  Initialize memory for the at&&.
----------------------------------------------------------------------------*/
void atDDRStart(void)
{
    atMportClassName = utSymCreate("atMport");
    atMportFunctionTable.className = atMportClassName;
    _atFirstFreeMport = atMportNull;
    _atMportHeap = utcHeap();
    buildMportBlock();
    _atFirstFreeContext = atContextNull;
    _atContextHeap = utcHeap();
    buildContextBlock();
    _atFirstFreeAttr = atAttrNull;
    _atAttrHeap = utcHeap();
    buildAttrBlock();
    atPortClassName = utSymCreate("atPort");
    atPortFunctionTable.className = atPortClassName;
    _atFirstFreePort = atPortNull;
    _atPortHeap = utcHeap();
    buildPortBlock();
    _atFirstFreeRoot = atRootNull;
    _atRootHeap = utcHeap();
    buildRootBlock();
    _atFirstFreeDevice = atDeviceNull;
    _atDeviceHeap = utcHeap();
    buildDeviceBlock();
    _atFirstFreePin = atPinNull;
    _atPinHeap = utcHeap();
    buildPinBlock();
}

/*----------------------------------------------------------------------------
  Free memory for the at&&.
----------------------------------------------------------------------------*/
void atDDRStop(void)
{
    utFreeHeap(_atMportHeap);
    utFreeHeap(_atContextHeap);
    utFreeHeap(_atAttrHeap);
    utFreeHeap(_atPortHeap);
    utFreeHeap(_atRootHeap);
    utFreeHeap(_atDeviceHeap);
    utFreeHeap(_atPinHeap);
}

